@@ -1,12 +1,18 @@
 __author__ = 'Rens Holmer'
-__version__ = '0.0.9'
+__version__ = '0.0.11'
 
 from .tree import Tree  # noqa
 from .sequence import (  # noqa
+    Alphabet,
+    DNA,
+    AMINO_ACID,
+    Sequence,
+    SequenceReader,
+    DNA,
+    AMINO_ACID,
     SequenceCollection,
     SequenceList,
     MultipleSequenceAlignment,
     SequenceAnnotation,
     SequenceInterval,
-    msa_plot,
 )
